"""
The topography module provides interfaces to global elevation models.
So far, only an interface to the `SRTM30
<https://dds.cr.usgs.gov/srtm/version2_1/SRTM30/srtm30_documentation.pdf>`_ data
set is provided, which has a resolution of 1 km.

Elevation data is downloaded on the fly but is cached to speed up subsequent
access. The interfaces uses the path pointed to by the :code:`TYPHON_DATA_PATH`
environment variable as data cache. This means that data is downloaded only
when they are not found in the cache.

.. note:: If :code:`TYPHON_DATA_PATH` is not set, the location of the file cache
    will be determined from the :code:`XDG_CACHE_HOME` environment variable and,
    if this is not defined, default to :`${HOME}/.typhon/topography`.

The module can be used in two ways: 
 1. by extracting the elevation data at native resolution
 2. by interpolating to elevation data at arbitrary locations

The two different use cases are described below.

Native resolution
-----------------

Extracting elevation data at native resolution for a given rectangular domain
 is done using the :code:`SRTM30.elevation` function. The function returns a
 tuple consisting of the latitude and longitude grids as well as the elevation
 data in meters.

.. code-block:: python

    lat_min = 50
    lon_min = 10
    lat_max = 60
    lon_max = 20
    lats, lons, z = SRTM30.elevation(lat_min, lon_min, lat_max, lon_max)

Interpolation to given coordinates
----------------------------------

Interpolation of the elevation data to arbitrary coordinates can be performed
using the :code:`interpolate` method. Interpolation uses nearest neighbor
interpolation and is implemented using a :code:`KDTree`.
Interpolating the SRTM30 data to given latitude and longitude grids can be done
as follows:

.. code-block:: python

    lat_min = 50
    lon_min = 10
    lat_max = 60
    lon_max = 20
    lats = np.linspace(lat_min, lat_max, 101)
    lons = np.linspace(lon_min, lon_max, 101)
    z = SRTM30.interpolate(lat, lont)
"""
import os
import shutil
import urllib
import zipfile

import numpy as np

import typhon
from typhon.environment import environ

_data_path = None

def _get_data_path():
    global _data_path
    if _data_path is None:
        if "TYPHON_DATA_PATH" in environ:
            _data_path = os.path.join(environ["TYPHON_DATA_PATH"], "topography")
        elif "XDG_CACHE_HOME" in environ:
            _data_path = environ["XDG_CACHE_HOME"]
        else:
            home = os.path.expandvars("~")
            _data_path = os.path.join(home, ".cache", "typhon", "topography")
        if not os.path.exists(_data_path):
            os.makedirs(_data_path)
    return _data_path

def _latlon_to_cart(lat, lon, R = typhon.constants.earth_radius):
    """
    Simple conversion of latitude and longitude to Cartesian coordinates.
    Approximates the Earth as sphere with radius :code:`R` and computes
    cartesian x, y, z coordinates with the center of the Earth as origin.

    Args:
        lat: Array of latitude coordinates.
        lon: Array of longitude coordinates.
        R: The radius to assume.
    Returns:
        Tuple :code:`(x, y, z)` of arrays :code:`x, y, z` containing the
        resulting x-, y- and z-coordinates.
    """
    lat = np.radians(lat)
    lon = np.radians(lon)
    x = R * np.cos(lat) * np.cos(lon)
    y = R * np.cos(lat) * np.sin(lon)
    z = R * np.sin(lat)
    return x, y, z

def _do_overlap(rect_1,
                rect_2):
    """
    Determines whether the two rectangles have overlap.

    Args:
        rect_1: Tuple :code:`(lat_min, lon_min, lat_max, lon_max) describing
                a rectangular tile.
        rect_2: Tuple :code:`(lat_min, lon_min, lat_max, lon_max) describing
                a rectangular tile.

    Returns:
        True if the two rectangles overlap.
    """
    lat_min_1, lon_min_1, lat_max_1, lon_max_1 = rect_1
    lat_min_2, lon_min_2, lat_max_2, lon_max_2 = rect_2
    lat_min = max(lat_min_1, lat_min_2)
    lon_min = max(lon_min_1, lon_min_2)
    lat_max = min(lat_max_1, lat_max_2)
    lon_max = min(lon_max_1, lon_max_2)
    return (lat_min < lat_max) and (lon_min < lon_max)

class SRTM30:
    """
    Interface to version 2.1 of SRTM30 digital elevation model.

    The data set has a resolution of about 1 km and covers all land masses
    except Antarctica.
    """
    _tile_height = 6000
    _tile_width = 4800
    _dlat = 50.0 / _tile_height
    _dlon = 40.0 / _tile_width

    _tiles = [("w180n90",  40, -180,  90, -140),
              ("w140n90",  40, -140,  90, -100),
              ("w100n90",  40, -100,  90,  -60),
              ("w060n90",  40,  -60,  90,  -20),
              ("w020n90",  40,  -20,  90,   20),
              ("e020n90",  40,   20,  90,   60),
              ("e060n90",  40,   60,  90,  100),
              ("e100n90",  40,  100,  90,  140),
              ("e140n90",  40,  140,  90,  180),
              ("w180n40", -10, -180,  40, -140),
              ("w140n40", -10, -140,  40, -100),
              ("w100n40", -10, -100,  40,  -60),
              ("w060n40", -10,  -60,  40,  -20),
              ("w020n40", -10,  -20,  40,   20),
              ("e020n40", -10,   20,  40,   60),
              ("e060n40", -10,   60,  40,  100),
              ("e100n40", -10,  100,  40,  140),
              ("e140n40", -10,  140,  40,  180),
              ("w180s10", -60, -180, -10, -140),
              ("w140s10", -60, -140, -10, -100),
              ("w100s10", -60, -100, -10,  -60),
              ("w060s10", -60,  -60, -10,  -20),
              ("w020s10", -60,  -20, -10,   20),
              ("e020s10", -60,   20, -10,   60),
              ("e060s10", -60,   60, -10,  100),
              ("e100s10", -60,  100, -10,  140),
              ("e140s10", -60,  140, -10,  180)]

    @staticmethod
    def get_tiles(lat_min, lon_min, lat_max, lon_max):
        """
        Get names of the tiles that contain the data of the given rectangular
        region of interest (ROI).

        Args:
            lat_min: The latitude of the lower left corner of the ROI
            lon_min: The longitude of the lower left corner of the ROI
            lat_max: The latitude of the upper right corner of the ROI
            lon_max: The longitude of the upper right corner of the ROI

        Return:

            List of tile names that contain the elevation data for the ROI.
        """
        lon_min = lon_min % 360
        if lon_min > 180:
            lon_min -= 360

        lon_max = lon_max % 360
        if lon_max > 180:
            lon_max -= 360


        fits = []
        for t in SRTM30._tiles:
            name, lat_min_1, lon_min_1, lat_max_1, lon_max_1 = t
            if _do_overlap((lat_min, lon_min, lat_max, lon_max),
                          (lat_min_1, lon_min_1, lat_max_1, lon_max_1)):
                fits += [name]
        return fits

    @staticmethod
    def get_bounds(name):
        """
        Get the bounds of tile with a given name.
        Args:
            name(str): The name of the tile.
        Returns:
            Tuple :code:(`lat_min`, `lon_min`, `lat_max`, `lon_max`) describing
            the bounding box of the tile with the given name.
        """
        tile = [t for t in SRTM30._tiles if t[0] == name][0]
        _, lat_min, lon_min, lat_max, lon_max = tile
        return lat_min, lon_min, lat_max, lon_max

    @staticmethod
    def get_grids(name):
        """
        Returns the latitude-longitude grid of the tile with the given name.

        Args:
            name(str): The name of the tile.
        Returns:
            Tuple :code:(`lat_grid`, `lon_grid`) containing the one dimensional
            latitude and longitude grids corresponding to the given tile.
        """
        lat_min, lon_min, lat_max, lon_max = SRTM30.get_bounds(name)
        start = lat_min + 0.5 * SRTM30._dlat
        stop = lat_max - 0.5 * SRTM30._dlat
        lat_grid = np.linspace(start, stop, SRTM30._tile_height)[::-1]

        start = lon_min + 0.5 * SRTM30._dlon
        stop = lon_max - 0.5 * SRTM30._dlon
        lon_grid = np.linspace(start, stop, SRTM30._tile_width)

        return lat_grid, lon_grid

    @staticmethod
    def get_native_grids(lat_min, lon_min, lat_max, lon_max):
        """
        Returns the latitude and longitude grid at native SRTM30 resolution
        that are included in the given rectangle.

        Args:
            lat_min: The latitude coordinate of the lower left corner.
            lon_min: The longitude coordinate of the lower left corner.
            lat_max: The latitude coordinate of the upper right corner.
            lon_max: The latitude coordinate of the upper right corner.
        Returns:
            Tuple :code:`(lats, lons)` of 1D-arrays containing the latitude
            and longitude coordinates of the SRTM30 data points within the
            given rectangle.
        """
        i = (90 - lat_max) / SRTM30._dlat
        i_max = np.trunc(i)
        if not i_max < i:
            i_max = i_max + 1
        i = (90 - lat_min) / SRTM30._dlat
        i_min = np.trunc(i)
        lat_grid = 90 + 0.5 * SRTM30._dlat - np.arange(i_max, i_min + 1) * SRTM30._dlat

        j = (lon_max + 180) / SRTM30._dlon
        j_max = np.trunc((lon_max + 180.0) / SRTM30._dlon)
        if not j_max < j:
            j_max = j_max - 1

        j_min = np.trunc((lon_min + 180.0) / SRTM30._dlon)
        lon_grid = -180 + 0.5 * SRTM30._dlon
        lon_grid += np.arange(j_min, j_max + 1) * SRTM30._dlon

        return lat_grid, lon_grid

    @staticmethod
    def download_tile(name):
        """
        This function will download and extract the tile with the given name.
        The data is stored in the path pointed to by the :code:`_data_path`
        attribute of the module.

        Args:
            name(str): The name of the tile to download.
        """
        base_url = "https://dds.cr.usgs.gov/srtm/version2_1/SRTM30"
        url = base_url + "/" + name + "/" + name + ".dem.zip"
        r = urllib.request.urlopen(url)

        filename = os.path.join(_get_data_path(), name + ".dem.zip")
        path = os.path.join(filename)
        with open(path, 'wb') as f:
            shutil.copyfileobj(r, f)

        # Extract zip file.
        with zipfile.ZipFile(filename, "r") as zip_ref:
            zip_ref.extractall(os.path.dirname(filename))

    @staticmethod
    def get_tile(name):
        """
        Get tile with the given name.

        Check the cache for the tile with the given name. If not found, the
        tile is download.

        Args:
            name(str): The name of the tile.
        """
        dem_file = os.path.join(_get_data_path(), (name + ".dem").upper())
        if not (os.path.exists(dem_file)):
            SRTM30.download_tile(name)
        y = np.fromfile(dem_file, dtype = np.dtype('>i2')).reshape(SRTM30._tile_height,
                                                                   SRTM30._tile_width)
        return y

    @staticmethod
    def get_tree(name):
        """
        Get KD-tree for the tile with the given name.

        Args:
            name(str): The name of the tile.
        """
        from pykdtree.kdtree import KDTree
        lat_grid, lon_grid = SRTM30.get_grids(name)
        lat_grid, lon_grid = np.meshgrid(lat_grid, lon_grid, indexing = "ij")
        x, y, z = _latlon_to_cart(lat_grid, lon_grid)
        X = np.concatenate([x.reshape(-1, 1, order = "C"),
                            y.reshape(-1, 1, order = "C"),
                            z.reshape(-1, 1, order = "C")], axis = 1)
        tree = KDTree(X.astype(np.float32))
        return tree

    @staticmethod
    def elevation(lat_min,
                  lon_min,
                  lat_max,
                  lon_max):
        """
        Return elevation data at native resolution in the a given rectangular
        domain.

        Args:
            lat_min(float): Latitude coordinate of the lower-left corner
            lon_min(float): Longitude coordinate of the lower-left corner.
            lat_max(float): Latitude coordinate of the upper-right corner
            lon_max(float): Longitude coordinate of the upper-right corner.

        """
        lats_d, lons_d = SRTM30.get_native_grids(lat_min,
                                                 lon_min,
                                                 lat_max,
                                                 lon_max)
        lat_min = lats_d.min() - 0.5 * SRTM30._dlat
        lat_max = lats_d.max() + 0.5 * SRTM30._dlat
        lon_min = lons_d.min() - 0.5 * SRTM30._dlon
        lon_max = lons_d.max() + 0.5 * SRTM30._dlon

        elevation = np.zeros(lats_d.shape + lons_d.shape)
        tiles = SRTM30.get_tiles(lat_min, lon_min, lat_max, lon_max)
        for t in tiles:
            dem = SRTM30.get_tile(t)
            lats, lons = SRTM30.get_grids(t)
            lat_min_s, lon_min_s, lat_max_s, lon_max_s = SRTM30.get_bounds(t)

            inds_lat = np.logical_and(lat_min <= lats, lats < lat_max)
            inds_lon = np.logical_and(lon_min <= lons, lons < lon_max)
            inds_s = np.logical_and(inds_lat.reshape(-1, 1),
                                    inds_lon.reshape(1, -1))

            inds_lat = np.logical_and(lat_min_s <= lats_d, lats_d < lat_max_s)
            inds_lon = np.logical_and(lon_min_s <= lons_d, lons_d < lon_max_s)
            inds_d = np.logical_and(inds_lat.reshape(-1, 1),
                                    inds_lon.reshape(1, -1))

            elevation[inds_d] = dem[inds_s]

        return lats_d, lons_d, elevation

    @staticmethod
    def interpolate(lats,
                    lons,
                    n_neighbors = 1):
        """
        Interpolate elevation data to the given coordinates.

        Uses KD-tree-based nearest-neighbor interpolation to interpolate
        the elevation data to arbitrary grids.

        Args:
            lats: Array containing latitude coordinates.
            lons: Array containing longitude coordinates.
            n_neighbors: Number of neighbors over which to average the elevation
                data.

        """
        lat_min = lats.min()
        lat_max = lats.max()
        lon_min = lons.min()
        lon_max = lons.max()
        tiles = SRTM30.get_tiles(lat_min, lon_min, lat_max, lon_max)

        elevation = np.zeros(lats.shape)
        for t in tiles:
            dem = SRTM30.get_tile(t).ravel()
            tree = SRTM30.get_tree(t)

            lat_min, lon_min, lat_max, lon_max = SRTM30.get_bounds(t)
            inds_lat = np.logical_and(lat_min <= lats, lats < lat_max)
            inds_lon = np.logical_and(lon_min <= lons, lons < lon_max)
            inds = np.logical_and(inds_lat, inds_lon)

            X = np.zeros((inds.sum(), 3))
            x, y, z = _latlon_to_cart(lats[inds], lons[inds])
            X[:, 0] = x
            X[:, 1] = y
            X[:, 2] = z

            _, neighbors = tree.query(np.asarray(X, np.float32), n_neighbors)
            if neighbors.size > 0:
                if len(neighbors.shape) > 1:
                    elevation[inds] = dem[neighbors].mean(axis = (1))
                else:
                    elevation[inds] = dem[neighbors]
        return elevation



from satpy import Scene
import xarray as xr

from .common import expects_file_info, FileHandler


class SatPy(FileHandler):
    """Wrapper for SatPy readers

    SatPy provides many readers for different satellite files. With this
    simple wrapper class, you can use them with typhon's FileSet without
    requiring any extra code!

    Warnings:
        At the moment, this wrapper works for retrieving fields with the
        same resolution only. Latitudes and longitudes are not retrieved.
    """

    def __init__(self, satpy_reader):
        super(SatPy, self).__init__()

        self.satpy_reader = satpy_reader

    @expects_file_info()
    def read(self, filename, fields=None, **kwargs):
        scene = Scene(
            reader=self.satpy_reader,
            filenames=[filename.path]
        )

        # If the user has not passed any fields to us, we load all per default.
        if fields is None:
            fields = scene.available_dataset_ids()

        # Load all selected fields
        scene.load(fields, **kwargs)

        if isinstance(fields[0], str):
            data_arrays = {field: scene.get(field) for field in fields}
        else:
            data_arrays = {field.name: scene.get(field) for field in fields}

        for name, array in data_arrays.items():
            array.name = name

        dataset = xr.merge(data_arrays.values())

        return dataset





# -*- coding: utf-8 -*-

"""Module for anything related to the electromagnetic spectrum.

This module imports typhon.physics.units.common and therefore
has a soft dependency on the pint units library.
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import warnings
import logging

import numpy
import scipy.interpolate

import numexpr
import pint
import xarray


from typhon import config
from typhon.arts import xml
from typhon.constants import (h, k, c)
from typhon.physics.units.common import (ureg, radiance_units)
from typhon.physics.units.tools import UnitsAwareDataArray as UADA


logger = logging.getLogger(__name__)


__all__ = [
    'FwmuMixin',
    'SRF',
    'planck_f',
    'specrad_wavenumber2frequency',
    'specrad_frequency_to_planck_bt',
    ]


class FwmuMixin:
    """Mixing for frequency/wavelength/wavenumber neutrality

    Best to use pint ureg quantities at all times.
    """
    _frequency = None
    _wavenumber = None
    _wavelength = None

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        try:
            self._frequency = value.to(ureg.Hz, "sp")
        except AttributeError:
            value = value * ureg.Hz
            self._frequency = value

        self._wavenumber = value.to(1 / ureg.centimeter, "sp")
        self._wavelength = value.to(ureg.metre, "sp")

    @property
    def wavenumber(self):
        return self._wavenumber

    @wavenumber.setter
    def wavenumber(self, value):
        try:
            self._wavenumber = value.to(1 / ureg.centimeter, "sp")
        except AttributeError:
            value = value * 1 / ureg.centimeter
            self._wavenumber = value

        self._frequency = value.to(ureg.Hz, "sp")
        self._wavelength = value.to(ureg.metre, "sp")

    @property
    def wavelength(self):
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value):
        try:
            self._wavelength = value.to(ureg.metre, "sp")
        except AttributeError:
            value = value * ureg.meter
            self._wavelength = value

        self._frequency = value.to(ureg.hertz, "sp")
        self._wavenumber = value.to(1 / ureg.centimeter, "sp")


class SRF(FwmuMixin):
    """Respresents a spectral response function

    TODO: representation of uncertainties
    """

    T_lookup_table = numpy.arange(0, 500.01, 0.05) * ureg.K
    lookup_table = None
    L_to_T = None

    def __init__(self, f, W):
        """Initialise SRF object.

        You can either initialise an SRF from scratch, or use the
        classmethod `fromArtsXML` to read it from a file.

        A toy example on initiating it from scratch:

        >>> from typhon.physics.units.common import ureg
        >>> from typhon.physics.units.em import SRF
        >>> srf = SRF(ureg.Quantity(numpy.array([200, 200.1, 200.2, 200.3, 200.4, 200.5]), 'GHz'), numpy.array([0, 0.5, 1, 1, 0.5, 0]))
        >>> R_300 = srf.blackbody_radiance(ureg.Quantity(300, 'K'))
        >>> print(R_300)
        [  3.63716781e-15] watt / hertz / meter ** 2 / steradian
        >>> print(R_300.to("K", "radiance", srf=srf))
        [ 300.] kelvin

        You can also pass in other spectroscopic units (wavenumber,
        wavelength) that will be converted internally to frequency:

        >>> srf = SRF(ureg.Quantity(numpy.array([10.8, 10.9, 11.0, 11.1, 11.2, 11.3]), 'um'), numpy.array([0, 0.5, 1, 1, 0.5, 0]))
        >>> R_300 = srf.blackbody_radiance(ureg.Quantity(numpy.atleast_1d(250), 'K'))
        >>> print(R_300)
        [  1.61922509e-12] watt / hertz / meter ** 2 / steradian
        >>> print(R_300.to("cm * mW / m**2 / sr", "radiance"))
        [ 48.54314703] centimeter * milliwatt / meter ** 2 / steradian
        >>> print(R_300.to("K", "radiance", srf=srf))
        [ 300.] kelvin

        :param ndarray f: Array of frequencies.  Can be either a pure
            ndarray, which will be assumed to be in Hz, or a ureg
            quantity.
        :param ndarray W: Array of associated weights.
        """

        try:
            self.frequency = f.to("Hz", "sp")
        except AttributeError:
            self.frequency = ureg.Quantity(f, "Hz")
        self.W = W

    def __repr__(self):
        """Get string representation.

        Uses centroid.
        """
        cf = self.centroid()
        if cf.to("Hz", "sp").m > 3e12:
            s = cf.to(ureg.um, "sp")
        else:
            s = cf.to(ureg.GHz, "sp")
        return "<{:s}: {:.4~}>".format(self.__class__.__name__, s)

    @classmethod
    def fromArtsXML(cls, sat, instr, ch):
        """Read SRF from ArtsXML files.

        Requires that in the TYPHONRC configuration file, the fields
        `srf_backend_f` and `srf_backend_response` in the section
        corresponding to instrument `instr` are defined to point to the
        respective files in ArtsXML format.  Within those definitions,
        {sat:s} will be substituted with the satellite name.  For example,
        in typhonrc, one might have:

        [hirs]
        srf_backend_response = /path/to/{sat}_HIRS.backend_channel_response.xml
        srf_backend_f = /path/to/{sat}_HIRS.f_backend.xml

        so that we can do:

        >>> srf = SRF.fromArtsXML("NOAA15", "hirs", 12)
        >>> R_300 = srf.blackbody_radiance(ureg.Quantity(atleast_1d(250), 'K')) 
        >>> print(R_300)
        [  2.13002925e-13] watt / hertz / meter ** 2 / steradian
        >>> print(R_300.to("cm * mW / m**2 / sr", "radiance")) 
        [ 6.38566704] centimeter * milliwatt / meter ** 2 / steradian

        Arguments:

            sat [str]

                Satellite name, such as 'NOAA15'

            instr [str]

                Instrument name, such as 'hirs'.

            ch [int]

                Channel number (start counting at 1).
        """
        cf = config.conf[instr]
        centres = xml.load(
            cf["srf_backend_f"].format(sat=sat))
        gfs = xml.load(
            cf["srf_backend_response"].format(sat=sat))
        freq = gfs[ch - 1].grids[0] + centres[ch - 1]
        response = gfs[ch - 1].data
        return cls(freq, response)

    @classmethod
    def fromRTTOV(cls, sat, instr, ch):
        """Read SRF from RTTOV format files.

        Requires that in the TYPHONRC configuration file, the fields
        'srf_rttov' in the section corresponding to instrument `instr` are
        defined to point to the respective files in RTTOV format.  Within
        those definitions, {sat:s} will be substituted with the satellite
        name and {ch:>02d} with the channel number.  For example, in
        typhonrc, we might have:

        [hirs]
        srf_rttov = /path/to/rtcoef_{sat:s}_hirs-shifted_src_ch{ch:>02d}.txt

        Arguments:

            sat [str]

                Satellite name, such as 'noaa-15'

            instr [str]

                Instrument name, such as 'hirs'.

            ch [int]

                Channel number (start counting at 1).
        """
        cf = config.conf[instr]
        M = numpy.loadtxt(cf["srf_rttov"].format(sat=sat, ch=ch),
                          skiprows=4)
        wn = ureg.Quantity(M[:, 0], 1/ureg.cm)
        W = M[:, 1]
        return cls(wn, W)

    def centroid(self):
        """Calculate centre frequency
        """
        return numpy.average(
            self.frequency, weights=self.W) * self.frequency.units

    def blackbody_radiance(self, T, spectral=True):
        """Calculate integrated radiance for blackbody at temperature T

        :param T: Temperature [K].  This can be either a python number, or
            a numpy ndarray, on a ureg quantity encompassing either.
        :param spectral: Parameter to control whether to return spectral
            radiance or radiance.  See self.integrate_radiances for
            details.

        Returns quantity ndarray with blackbody radiance in desired unit.
        Note that this is an ndarray with dimension (1,) even if you
        passin a scalar.
        """
        try:
            T = T.to("K")
        except AttributeError:
            T = ureg.Quantity(T, "K")
        T = ureg.Quantity(numpy.atleast_1d(T), T.u)
        # fails if T is multidimensional
        shp = T.shape
        return self.integrate_radiances(
            self.frequency, planck_f(
                self.frequency[numpy.newaxis, :],
                T.reshape((-1,))[:, numpy.newaxis]),
                spectral=spectral).reshape(shp)

    def make_lookup_table(self):
        """Construct lookup table radiance <-> BT

        To convert a channel radiance to a brightness temperature,
        applying the (inverse) Planck function is not correct, because the
        Planck function applies to monochromatic radiances only.  Instead,
        to convert from radiance (in W m^-2 sr^-1 Hz^-1) to brightness
        temperature, we use a lookup table.  This lookup table is
        constructed by considering blackbodies at a range of temperatures,
        then calculating the channel radiance.  This table can then be
        used to get a mapping from radiance to brightness temperature.

        This method does not return anything, but fill self.lookup_table.
        """
        self.lookup_table = numpy.zeros(
            shape=(2, self.T_lookup_table.size), dtype=numpy.float64)
        self.lookup_table[0, :] = self.T_lookup_table
        self.lookup_table[1, :] = self.blackbody_radiance(self.T_lookup_table)
        self.L_to_T = scipy.interpolate.interp1d(self.lookup_table[1, :],
                                                 self.lookup_table[0, :],
                                                 kind='linear',
                                                 bounds_error=False,
                                                 fill_value=(0, 2000))

    def integrate_radiances(self, f, L, spectral=True):
        """From a spectrum of radiances and a SRF, calculate channel (spectral) radiance

        The spectral response function may not be specified on the same grid
        as the spectrum of radiances.  Therefore, this function interpolates
        the spectral response function onto the grid of the radiances.  This
        is less bad than the reverse, because a spectral response function
        tends to be more smooth than a spectrum.

        **Approximations:**

        * Interpolation of spectral response function onto frequency grid on
          which radiances are specified.

        :param ndarray f: Frequencies for spectral radiances [Hz]
        :param ndarray L: Spectral radiances [W m^-2 sr^-1 Hz^-1].
            Can be in radiance units of various kinds.  Make sure this is
            consistent with the spectral response function.
            Innermost dimension must correspond to frequencies.
        :param bool spectral: If true, return spectral radiance
            [W m^-2 sr^-1 Hz^-1].  If false, return radiance [W m^-2
            sr^-1].  Defaults to True.
        :returns: Channel (spectral) radiance according to 'spectral'
        """
        # Interpolate onto common frequency grid.  The spectral response
        # function is more smooth so less harmed by interpolation, so I
        # interpolate the SRF.
        fnc = scipy.interpolate.interp1d(
            self.frequency, self.W, bounds_error=False, fill_value=0.0)
        #w_on_L_grid = fnc(f) * (1 / ureg.Hz)
        w_on_L_grid = ureg.Quantity(fnc(f), ureg.dimensionless)# * (1 / ureg.Hz)

        df = ureg.Quantity(numpy.diff(f), f.u)
        w1p = w_on_L_grid[1:]
        L1p = L[:, 1:]
        # ch_BT = (w_on_L_grid * L_f).sum(-1) / (w_on_L_grid.sum())
        # due to numexpr limitation, do sum seperately
        # and due to numexpr bug, explicitly consider zero-dim case
        #     see https://github.com/pydata/numexpr/issues/229
        if L.shape[0] == 0:
            ch_rad = numpy.empty(dtype="f8", shape=L.shape[:-1])
        else:
#            ch_rad = numexpr.evaluate("sum(w_on_L_grid * L, {:d})".format(
            ch_rad = numexpr.evaluate("sum(w1p * L1p * df, {:d})".format(
                L.ndim - 1))
        ch_rad = ureg.Quantity(ch_rad, w1p.u * L1p.u * df.u)
        if spectral:
            return ch_rad / (w1p*df).sum()
        else:
            return ch_rad

    def channel_radiance2bt(self, L):
        """Convert channel radiance to brightness temperature

        Using the lookup table, convert channel radiance to brightness
        temperature.  Will construct lookup table on first call.

        Typhon also registers a pint context “radiance” which can be used
        to convert between radiance units and brightness temperature (even
        though this is a different quantity), for example, by using
        L.to("K", "radiance", srf=srf)

        :param L: Radiance [W m^-2 sr^-1 Hz^-1] or compatible
        """
        if self.lookup_table is None:
            self.make_lookup_table()
        return self.L_to_T(
            L.to(ureg.W / (ureg.m**2 * ureg.sr * ureg.Hz),
                 "radiance")) * ureg.K

    def estimate_band_coefficients(self, sat=None, instr=None, ch=None,
            include_shift=True):
        """Estimate band coefficients for fast/explicit BT calculations

        In some circumstances, a fully integrated SRF may be more
        expensive than needed.  We can then choose an effective wavelength lambda_c
        along with coefficients alpha, beta such that instead of integrating, we
        estimate R = B(lambda*, T*), with T* = alpha + beta · T_B and lambda* a wavelength
        which may be close to the centroid lambda_c (but there is no
        guarantee).  Such an approximation eliminates the explicit use of
        an integral which can make analysis easier.

        Returns:

            alpha (float): Offset in approximation for T*
            beta (float): Slope in approximation for T*
            lambda_eff (float): Effective wavelength
            delta_alpha (float): Uncertainty in alpha
            delta_beta (float): Uncertainty in beta
            delta_lambda_eff (float): Uncertainty in lambda_eff
        """

        warnings.warn("Obtaining band coefficients from file", UserWarning)
        srcfile = config.conf[instr]["band_file"].format(sat=sat)
        if include_shift:
            rxp = r"(.{5,6})_ch(\d\d?)_shift([+-]\d+)pm\.nc\s+([\d.]+)\s+(-?[\de\-.]+)\s+([\d.]+)"
            dtp = [("satname", "S6"), ("channel", "u1"), ("shift", "i2"),
                   ("centre", "f4"), ("alpha", "f4"), ("beta", "f4")]
            M = numpy.fromregex(srcfile, rxp, dtp).reshape(19, 7)
            dims = ("channel", "shiftno")
            ds = xarray.Dataset(
                {"centre": (dims, M["centre"]),
                 "alpha": (dims, M["alpha"]),
                 "beta": (dims, M["beta"]),
                 "shift": (dims, M["shift"])},
                coords = {"channel": M["channel"][:, 0]})

        else:
            rxp = r"(.{5,6})_ch(\d\d?)_rttov\.nc\s+([\d.]+)\s+(-?[\de\-.]+)\s+([\d.]+)"
            dtp = [("satname", "S6"), ("channel", "u1"), 
                   ("centre", "f4"), ("alpha", "f4"), ("beta", "f4")]
            M = numpy.fromregex(srcfile, rxp, dtp)
            dims = ("channel",)
            ds = xarray.Dataset(
                {"centre": (dims, M["centre"]),
                 "alpha": (dims, M["alpha"]),
                 "beta": (dims, M["beta"])},
                coords = {"channel": M["channel"]})
        ds = ds.sel(channel=ch)

        if include_shift:
            ds0 = ds.sel(shiftno=0) # varies 1.1 – 15.2 nm depending on channel
        else:
            ds0 = ds
        lambda_c = UADA(ds0["centre"], attrs={"units": "1/cm"})
        alpha = UADA(ds0["alpha"], attrs={"units": "K"})
        beta = UADA(ds0["beta"], attrs={"units": "1"})

        if include_shift:
            delta_ds = ds.sel(shiftno=1) - ds0
            delta_lambda_c = abs(UADA(delta_ds["centre"], attrs={"units": "1/cm"}))
            delta_alpha = abs(UADA(delta_ds["alpha"], attrs={"units": "K"}))
            delta_beta = abs(UADA(delta_ds["beta"], attrs={"units": "1"}))
        else:
            # workaround for https://github.com/pydata/xarray/issues/2542
            # by passing to UADA
            delta_alpha = UADA(xarray.zeros_like(alpha))
            delta_alpha.attrs["units"] = "K"
            delta_beta = UADA(xarray.zeros_like(alpha))
            delta_beta.attrs["units"] = "1"
            delta_lambda_c = UADA(xarray.zeros_like(alpha))
            delta_lambda_c.attrs["units"] = "1/cm"

        return (alpha, beta, lambda_c, delta_alpha, delta_beta, delta_lambda_c)

    # Methods returning new SRFs with some changes
    def shift(self, amount):
        """Get new SRF, shifted by <amount>

        Return a new SRF, shifted by <amount> Hz.  The shape of the SRF is
        retained.

        Arguments:

            Quantity amount: Distance to shift SRF
        """
        return self.__class__(self.frequency.to(
            amount.u, "sp") + amount, self.W)

    def as_dataarray(self, coordinate):
        """Return xarray.DataArray object.

        Coordinate can be "wavelength" (which will be in m), "frequency"
        (which will be in Hz), or "wavenumber" (which will be in 1/cm).
        """

        return xarray.DataArray(self.W, dims=(coordinate,),
            coords={coordinate: getattr(self, coordinate)}, name="SRF")

_specrad_freq = ureg.W / (ureg.m**2 * ureg.sr * ureg.Hz)


def planck_f(f, T):
    """Planck law expressed in frequency.

    If more than 10⁵ resulting radiances, uses numexpr.

    :param f: Frequency.  Quantity in [Hz]
    :param T: Temperature.  Quantity in [K]
    """
#    try:
#        f = f.astype(numpy.float64)
#    except AttributeError:
#        pass
    if (f.size * T.size) > 1e5:
        return numexpr.evaluate("(2 * h * f**3) / (c**2) * "
                                "1 / (exp((h*f)/(k*T)) - 1)") * (
                                    radiance_units["si"])
    return ((2 * ureg.h * f**3) / (ureg.c ** 2) *
            1 / (numpy.exp(((ureg.h * f) / (ureg.k * T)).to("1")) - 1)).to(
                ureg.W / (ureg.m**2 * ureg.sr * ureg.Hz))


def specrad_wavenumber2frequency(specrad_wavenum):
    """Convert spectral radiance from per wavenumber to per frequency

    :param specrad_wavenum: Spectral radiance per wavenumber
         [W·sr^{-1}·m^{-2}·{m^{-1}}^{-1}]
    :returns: Spectral radiance per frequency [W⋅sr−1⋅m−2⋅Hz−1]
    """

    if not isinstance(specrad_wavenum, pint.quantity._Quantity):
        specrad_wavenum = specrad_wavenum * ureg.W / (
            ureg.m**2 * ureg.sr * (1 / ureg.m))

    return (specrad_wavenum / ureg.c).to(ureg.W / (
        ureg.m**2 * ureg.sr * ureg.Hz))


def specrad_frequency_to_planck_bt(L, f):
    """Convert spectral radiance per frequency to brightness temperature

    This function converts monochromatic spectral radiance per frequency
    to Planck brightness temperature.  This is calculated by inverting the
    Planck function.

    Note that this function is NOT correct to estimate polychromatic
    brightness temperatures such as channel brightness temperatures.  For
    this, you need the spectral response function — see the SRF class.

    :param L: Spectral radiance [W m^-2 sr^-1 Hz^-1]
    :param f: Corresponding frequency [Hz]
    :returns: Planck brightness temperature [K].
    """

    # f needs to be double to prevent overflow
    f = numpy.asarray(f).astype(numpy.float64)
    if L.size > 1500000:
        logger.debug("Doing actual BT conversion: {:,} spectra * {:,} "
                     "frequencies = {:,} radiances".format(
                         L.size // L.shape[-1], f.size, L.size))
    # BT = (h * f) / (k * numpy.log((2*h*f**3)/(L * c**2) + 1))
    BT = numexpr.evaluate("(h * f) / (k * log((2*h*f**3)/(L * c**2) + 1))")
    BT = numpy.ma.masked_invalid(BT)
    if L.size > 1500000:
        logger.debug("(done)")
    return ureg.Quantity(BT, ureg.K)



"""

The `workspace` subpackage provides an interactive interface to ARTS and
can be used to directly call workspace methods and agendas as well as
access and manipulate workspace variables.

Setup
-----

The python interface needs access to the :code:`libarts_api.so` shared library, which
is located in the `src` subfolder of the ARTS build tree. The interface expects
the location of the ARTS build tree to be provided in the ``ARTS_BUILD_PATH``
environment variable. For the :code:`libarts_api.so` to be built, ARTS needs to be
configured with :code:`-DENABLE_C_API=1`:

.. code-block:: bash

    cmake -DENABLE_C_API=1 ... /path/to/arts

The Workspace Class
-------------------

The main functionality of the interface is implemented by the Workspace class.
A Workspace object represents an ongoing ARTS simulation and is used to execute
controlfiles and workspace methods and access workspace variables

>>> from typhon.arts.workspace import Workspace
>>> ws = Workspace()

Executing Controlfiles
----------------------

Controlfiles can be executed on a workspace using the
:func:`~Workspace.execute_controlfile` method.

>>> ws.execute_controlfile("general.arts")

The search path for controlfiles is the current directory plus the paths
provided in the environment variable ``ARTS_INCLUDE_PATH``. Controlfiles
are parsed only once and then cached. Thus, if a controlfile is modified
after it has been loaded, these changes will only be effective after a
restart of the python runtime.

Calling Workspace Methods
-------------------------

ARTS workspace methods are available as member functions of each Workspace
object:

>>> ws.AtmosphereSet1D()
>>> ws.IndexSet(ws.stokes_dim, 1)

Arguments can be passed to a workspace function in three ways:

    1. As workspace variables using the attributes of the
       workspace object, such as :code:`ws.stokes_dim` in the
       example above.
    2. Using one of the symbolic variables in
       :mod:`typhon.arts.workspace.variables`
    3. Passing supported python objects directly

Arguments to a WSM can be passed using either positional or named arguments or both. If
positional arguments are provided at least all generic output and generic input arguments
must be given in order.

>>> ws.VectorNLogSpace(ws.p_grid, 361, 500e2, 0.1 )

Keyword arguments to define generic input and output arguments. Available keywords are the
names of the generic outputs and inputs defined in methods.cc.

>>> ws.abs_speciesSet(species=[ "O3" ])

Calls to supergeneric functions are resolved by the interface.

Workspace Variables
-------------------

Symbolic representation of all workspace variables are available in the typhon.arts.workspace.variables
module as module attributes. The purpose of these is to be passed to workspace functions as placeholders
for variables in the workspace.

Variable objects can be associated to a workspace, which is the case for variables accessed as attributes
of a workspace, i.e. using for example:

>>> ws.y

If that is the case their value can be accessed using the value() member function. In order
to print out a textual representation of the value, the print() member function can be used,
which will call the corresponding Print() ARTS WSV.

Workspace variables of the groups Vector, Matrix and Tensor with an associated workspace
implement the numpy array interface and can therefore be used just as any other numpy array.
In some cases, however, it may be necessary to explicitly create a view on the array using
numpy.asarray.

"""

import logging

from typhon.environment import environ


logger = logging.getLogger(__name__)

if environ.get('ARTS_BUILD_PATH') is None:
    logger.warning(
        "ARTS_BUILD_PATH environment variable required to locate ARTS API.")
else:
    from typhon.arts.workspace.workspace import Workspace, arts_agenda, Include
    from typhon.arts.workspace.variables import WorkspaceVariable
    from typhon.arts.workspace.methods   import WorkspaceMethod
    from typhon.arts.workspace.api       import arts_include_path   \
                                                , include_path_push \
                                                , data_path_push    \
                                                , data_path_pop     \




"""
The workspace submodule.

Contains the Workspace which implements the main functionality of the ARTS interface.
Users should only have to use this class to interact with ARTS.

Attributes:
     imports(dict): Dictionary of parsed controlfiles. This is kept to ensure to avoid
                    crashing of the ARTS runtime, when a file is parsed for the second time.

"""
import ctypes as c
import logging
import numpy  as np

import ast
from   ast      import iter_child_nodes, parse, NodeVisitor, Call, Attribute, Name, \
                       Expression, Expr, FunctionDef, Starred, Module, expr
from   inspect  import getsource, getclosurevars
from contextlib import contextmanager
from copy       import copy
from functools  import wraps
import os

from typhon.arts.workspace.api       import arts_api, VariableValueStruct, \
                                            data_path_push, data_path_pop, \
                                            include_path_push, include_path_pop
from typhon.arts.workspace.methods   import WorkspaceMethod, workspace_methods
from typhon.arts.workspace.variables import WorkspaceVariable, group_names, group_ids, \
                                            workspace_variables
from typhon.arts.workspace.agendas   import Agenda
from typhon.arts.workspace import variables as V
from typhon.arts.workspace.output import CoutCapture
from typhon.arts.workspace.utility import unindent

imports = dict()


logger = logging.getLogger(__name__)

################################################################################
# ARTS Agenda Macro
################################################################################

class Include:
    """Simple helper class to handle INCLUDE statements in agenda definitions.

    Attributes:

        agenda: The included controlfile or agenda as
            typhon.arts.workspace.agenda.Agenda object.
    """
    def __init__(self, agenda):
        """ Create include from argument.

        Args:

            agenda (str, Agenda): Argument to the INCLUDE statement. This can
                either be a string or an Agenda object.
        """
        if type(agenda) == str:
            if not agenda in imports:
                self.agenda = Agenda.parse(agenda)
                imports[agenda] = self.agenda
            else:
                self.agenda = imports[agenda]
        elif type(agenda) == Agenda:
            self.agenda = agenda
        else:
            raise Exception("agenda argument must be either a controlfile"
            " name or a typhon.arts.workspace.agenda.Agenda object.")

def arts_agenda(func):
    """
    Parse python method as ARTS agenda

    This decorator can be used to define ARTS agendas using python function syntax.
    The function should have one arguments which is assumed to be a Workspace instance.
    All expressions inside the function must be calls to ARTS WSMs. The result is an
    Agenda object that can be used to copied into a named ARTS agenda

    Example:

    >>> @arts_agenda
    >>> def inversion_iterate_agenda(ws):
    >>>     ws.x2artsStandard()
    >>>     ws.atmfields_checkedCalc()
    >>>     ws.atmgeom_checkedCalc()
    >>>     ws.yCalc()
    >>>     ws.VectorAddVector(ws.yf, ws.y, ws.y_baseline)
    >>>     ws.jacobianAdjustAfterIteration()
    >>>
    >>> ws.Copy(ws.inversion_iterate_agenda, inversion_iterate_agenda)
    """

    source = getsource(func)
    source = unindent(source)
    ast = parse(source)

    func_ast = ast.body[0]
    if not type(func_ast) == FunctionDef:
        raise Exception("ARTS agenda definition can only decorate function definiitons.")

    args = func_ast.args.args

    try:
        arg_name = func_ast.args.args[0].arg
    except:
        raise Exception("Agenda definition needs workspace arguments.")

    ws = Workspace(0)

    context = copy(func.__globals__)
    context.update({arg_name : ws})
    # Add resolved non-local variables from closure.
    nls, _, _, _ = getclosurevars(func)
    context.update(nls)

    #
    # Helper functions
    #

    callback_body = []
    def callback_make_fun(body):
        """
        Helper function that creates a wrapper function around
        python code to be executed withing an ARTS agenda.
        """
        m = Module(body)

        def callback(ptr):
            try:
                context[arg_name].ptr = ptr
                eval(compile(m , "<unknown>", 'exec'), context)
            except Exception as e:
                logger.error(r"Exception in Python callback:\n", e)
            context[arg_name].ptr = None

        callback_body = []
        return callback

    def eval_argument(expr):
        """
        Evaluate argument of workspace method call.
        """
        if not hasattr(expr, "lineno"):
            setattr(expr, "lineno", 0)
        return eval(compile(Expression(expr), "<unknown>", 'eval'), context)

    # Create agenda
    a_ptr = arts_api.create_agenda(func.__name__.encode())
    agenda = Agenda(a_ptr)

    illegal_statement_exception = Exception(
        "Agenda definitions may only contain calls to WSMs of the"
        "workspace argument " + arg_name + " or INCLUDE statements.")

    #
    # Here the body of the function definition is traversed. Cases
    # that are treated specieal are INCLUDE statements and calls
    # of workspace methods. Remaining statements are accumulated
    # in callback_body and then added to the agenda as a single callback.
    #

    for e in func_ast.body:
        if not isinstance(e, Expr):
            callback_body += [e]
            continue
        else:
            call = e.value

        if not isinstance(call, Call):
            callback_body += [e]
            continue

        # Include statement
        if type(call.func) == Name:
            if not call.func.id == "INCLUDE":
                callback_body += [e]
            else:
                args = []
                for a in call.args:
                    args.append(eval_argument(a))
                    include = Include(*args)

                    if len(callback_body) > 0:
                        agenda.add_callback(callback_make_fun(callback_body))
                        callback_body = []

                    arts_api.agenda_append(agenda.ptr, include.agenda.ptr)
        else:
            att  = call.func.value
            if not att.id == arg_name:
                callback_body += [e]
                continue

            # Extract method name.
            name = call.func.attr

            # m is not a workspace method
            if not name in workspace_methods:
                callback_body += [e]
                continue

            # m is a workspace method.
            m  = workspace_methods[name]

            args = [ws, m]

            for a in call.args:
                # Handle starred expression
                if type(a) == Starred:
                    bs = eval_argument(a.value)
                    for b in bs:
                        args.append(b)
                    continue

                args.append(eval_argument(a))

            # Extract keyword arguments
            kwargs = dict()
            for k in call.keywords:
                kwargs[k.arg] = eval(
                    compile(Expression(k.value), "<unknown>", 'eval'),
                    context)

            # Add function to agenda
            if len(callback_body) > 0:
                agenda.add_callback(callback_make_fun(callback_body))
                callback_body = []

            agenda.add_method(*args, **kwargs)

    # Check if there's callback code left to add to the agenda.
    if len(callback_body) > 0:
        agenda.add_callback(callback_make_fun(callback_body))
        callback_body = []

    return agenda


################################################################################
# Workspace Method Wrapper Class
################################################################################
class WSMCall:
    """
    Wrapper class for workspace methods. This is necessary to be able to print
    the method doc as __repr__, which doesn't work for python function objects.

    Attributes:

        ws: The workspace object to which the method belongs.
        m:  The WorkspaceMethod object

    """
    def __init__(self, ws, m):
        self.ws = ws
        self.m  = m
        self.__doc__  = m.__doc__

    def __call__(self, *args, **kwargs):
        self.m.call(self.ws, *args, **kwargs)

    def __repr__(self):
        return repr(self.m)

################################################################################
# The Workspace Class
################################################################################
class Workspace:
    """
    The Workspace class represents an ongoing ARTS simulation. Each Workspace object
    holds its own ARTS workspace and can be used to execute ARTS workspace methods or
    access workspace variables.

    All workspace methods taken from workspace_methods in the methods module are added
    as attributed on creation and are thus available as class methods.

    Attributes:

        ptr(ctypes.c_void_p): object pointing to the ArtsWorkspace instance of the
            ARTS C API
        _vars(dict): Dictionary holding local variables that have been created
            interactively using the one of Create ARTS WSMs.


    """
    def __init__(self, verbosity=1, agenda_verbosity=0):
        """
        The init function just creates an instance of the ArtsWorkspace class of the
        C API and sets the ptr attributed to the returned handle.

        It also adds all workspace methods as attributes to the object.

        Parameters:
            verbosity (int): Verbosity level (0-3), 1 by default
            agenda_verbosity (int): Verbosity level for agendas (0-3),
                                    0 by default
        """

        self.__dict__["_vars"] = dict()
        self.ptr = arts_api.create_workspace(verbosity, agenda_verbosity)
        self.workspace_size = arts_api.get_number_of_variables()
        for name in workspace_methods:
            m = workspace_methods[name]
            setattr(self, m.name, WSMCall(self, m))
        self.__verbosity_init__()

    def __del__(self):
        """
        Cleans up the C API.
        """
        if not self.ptr is None:
            if not arts_api is None:
                arts_api.destroy_workspace(self.ptr)

    def __getstate__(self):
        return None

    def __setstate__(self):
        pass

    def __verbosity_init__(self):
        """
        Executes verbosityInit WSM directly through the ARTS api to suppress
        output.
        """
        wsm = workspace_methods["verbosityInit"]
        (m_id, args_out, args_in, ts) = wsm._parse_output_input_lists(self, [], {})
        arg_out_ptr = c.cast((c.c_long * len(args_out))(*args_out),
                             c.POINTER(c.c_long))
        arg_in_ptr  = c.cast((c.c_long * len(args_in))(*args_in),
                            c.POINTER(c.c_long))
        with CoutCapture(self, silent = True):
            e_ptr = arts_api.execute_workspace_method(self.ptr, m_id, len(args_out),
                                                      arg_out_ptr, len(args_in), arg_in_ptr)
        for t in ts[::-1]:
            t.erase()

    def create_variable(self, group, name):
        """
        Create a workspace variable.

        Args:

            group: The group name of the variable to create.

            name: The name of the variable to create. If None, the
            ARTS API will assign a unique name.

        """
        if not name is None:
            name = name.encode()

        group_id = group_ids[group]
        ws_id    = arts_api.add_variable(self.ptr, group_id, name)
        v        = arts_api.get_variable(ws_id)
        wsv      = WorkspaceVariable(ws_id,
                                     v.name.decode(),
                                     group_names[group_id],
                                     "User defined variable.",
                                     self)
        self._vars[wsv.name] = wsv
        return wsv

    def add_variable(self, var):
        """
        This will try to copy a given python variable to the ARTS workspace and
        return a WorkspaceVariable object representing this newly created
        variable.

        Types are natively supported by the C API are int, str, [str], [int], and
        numpy.ndarrays. These will be copied directly into the newly created WSV.

        In addition to that all typhon ARTS types the can be stored to XML can
        be set to a WSV, but in this case the communication will happen through
        the file systs (cf. WorkspaceVariable.from_typhon).

        The user should not have to call this method explicitly, but instead it
        is used by the WorkspaceMethod call function to transfer python
        variable arguments to the ARTS workspace.

        Args:
            var: Python variable of type int, str, [str], [int] or np.ndarray
            which should be copied to the workspace.
        """
        if type(var) == WorkspaceVariable:
            return var

        # Create WSV in ARTS Workspace
        group = group_names[WorkspaceVariable.get_group_id(var)]
        wsv = self.create_variable(group, None)

        # Set WSV value using the ARTS C API
        s  = VariableValueStruct(var)
        if s.ptr:

            e = arts_api.set_variable_value(self.ptr, wsv.ws_id, wsv.group_id, s)
            if e:
                arts_api.erase_variable(self.ptr, wsv.ws_id, wsv.group_id)
                raise Exception("Setting of workspace variable through C API "
                                " failed with  the " + "following error:\n"
                                + e.decode("utf8"))
        # If the type is not supported by the C API try to write the type to XML
        # and read into ARTS workspace.
        else:
            try:
                wsv.from_typhon(var)
            except:
                raise Exception("Could not add variable since + "
                                + str(type(var)) + " is neither supported by "
                                + "the C API nor typhon XML IO.")
        self._vars[wsv.name] = wsv
        return wsv

    def __dir__(self):
        return {**self._vars, **workspace_variables, **self.__dict__}

    def __getattr__(self, name):
        """ Lookup the given variable in the local variables and the ARTS workspace.

        Args:
            name(str): Name of the attribute (variable)

        Raises:
            ValueError: If the variable is not found.
        """

        group_id = None
        if name in self._vars:
            var = self._vars[name]
            var.update()
            return var
        else:
            i = arts_api.lookup_workspace_variable(name.encode())
            if i < 0:
                raise AttributeError("No workspace variable " + str(name) + " found.")
            vs = arts_api.get_variable(i)
            group_id    = vs.group
            description = vs.description.decode("utf8")

        # Get its symbolic representation
        wsv = WorkspaceVariable(i, name, group_names[group_id], description, self)
        return wsv

    def __setattr__(self, name, value):
        """ Set workspace variable.

        This will lookup the workspace variable name and try to set it to value.

        Args:
            name(str):  Name of the attribute (variable)
            value(obj): The value to set the workspace variable to.

        Raises:
            ValueError: If the variable is not found or if value cannot uniquely converted to
            a value of a workspace variable.
        """
        try:
            v = self.__getattr__(name)
        except:
            self.__dict__[name] = value
            return None

        # Handle empty list or None values.
        if value is None or (isinstance(value, list) and not value):
            arts_api.set_variable_value(self.ptr, v.ws_id, v.group_id,
                                        VariableValueStruct.empty())
            return None

        if type(value) == Agenda:
            arts_api.set_variable_value(self.ptr, v.ws_id, v.group_id,
                                        VariableValueStruct(value))
            return None

        t = self.add_variable(value)

        if not t.group_id == v.group_id:
            raise Exception("Incompatible groups: Workspace variable " + name +
                            " of group " + group_names[v.group_id] + " and value " + str(value)
                            + " of group " + group_names[t.group_id] + ".")

        self.Copy(v, t)

        # Remove t only if it wasn't an existing WSV already before.
        if not type(value) == WorkspaceVariable:
            t.erase()

    def execute_agenda(self, agenda):
        """ Execute agenda on workspace.

        Args:

            agenda (typhon.arts.workspace.agenda.Agenda): Agenda object to execute.

        Raises:

            ValueError: If argument is not of type typhon.arts.workspace.agenda.Agenda
        """

        value_error = ValueError("Argument must be of type agenda.")
        if not type(agenda) is Agenda:
            raise value_error

        include_path_push(os.getcwd())
        data_path_push(os.getcwd())

        agenda.execute(self)

        include_path_pop()
        data_path_pop()

    def execute_controlfile(self, name):
        """ Execute controlfile or agenda on workspace.

        This method looks recursively for a controlfile with the given name in the current
        directory and the arts include path. If such a file has been found it will be parsed
        and executed on the workspace.

        Args:

            name(str): Name of the controlfile

        Raises:

            Exception: If parsing of the controlfile fails.

        Returns:

            The controlfile as parsed Agenda object.

        """

        if not name in imports:
            agenda = Agenda.parse(name)
            imports[name] = agenda
        else:
            agenda = imports[name]

        self.execute_agenda(agenda)

        return agenda



"""Functions operating on arrays."""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
#
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import numpy as np
import scipy.stats


def localmin(arr):
    """Find local minima for 1-D array

    Given a 1-dimensional numpy.ndarray, return the locations of any local
    minimum as a boolean array.  The first and last item are always
    considered False.

    Arguments:

        localmin (numpy.ndarray): 1-D ndarray for which to find local
            minima.  Should have a numeric dtype.

    Returns:

        numpy.ndarray with dtype `bool`.  True for any element that is
        strictly smaller than both neighbouring elements.  First and last
        element are always False.
    """

    localmin = np.hstack(
        (False, (arr[1:-1] < arr[0:-2]) & (arr[1:-1] < arr[2:]), False)
    )

    return localmin


def limit_ndarray(M, limits):
    """Select elements from structured ndarray based on value ranges

    This function filters a structured ndarray based on ranges defined for
    zero or more fields.  For each field f with limits (lo, hi), it will
    select only those elements where lo<=X[f]<hi.

    >>> X = array([(2, 3), (4, 5), (8, 2), (5, 1)],
                   dtype=[("A", "i4"), ("B", "i4")])

    >>> print(limit_ndarray(X, {"A": (2, 5)}))
    [(2, 3) (4, 5)]

    >>> X = array([([2, 3], 3), ([4, 6], 5), ([8, 3], 2), ([5, 3], 1)],
                   dtype=[("A", "i4", 2), ("B", "i4")])

    >>> print(limit_ndarray(X, {"A": (2, 5, "all")}))
    [([2, 3], 3)]

    Arguments:

        M (numpy.ndarray): 1-D structured ndarray

        limits (dict): Dictionary with limits.  Keys must correspond to
            fields in M.  If this is a scalar field
            (`M.dtype[field].shape==()`), values are tuples (lo, hi).
            If this is a multidimensional field, values are tuples (lo,
            hi, mode), where mode must be either `all` or `any`.
            Values in the range [lo, hi) are retained, applying all or any
            when needed.

    Returns:

        ndarray subset of M.  This is a view, not a copy.
    """

    selection = np.ones(shape=M.shape, dtype="?")

    for (field, val) in limits.items():
        ndim = len(M.dtype[field].shape)
        if ndim == 0:
            (lo, hi) = val
            selection = selection & (M[field] >= lo) & (M[field] < hi)
        else:
            (lo, hi, mode) = val
            lelo = M[field] >= lo
            sthi = M[field] < hi
            while lelo.ndim > 1:
                lelo = getattr(lelo, mode)(-1)
                sthi = getattr(sthi, mode)(-1)
            selection = selection & lelo & sthi

    return M[selection]


def parity(v):
    """Vectorised parity-checking.

    For any ndarray with an nd.integer dtype, return an equally shaped
    array with the bit parity for each element.

    Arguments:

        v (numpy.ndarray): Array of integer dtype

    Returns:

        ndarray with uint8 dtype with the parity for each value in v
    """

    v = v.copy()  # don't ruin original
    parity = np.zeros(dtype=">u1", shape=v.shape)
    while v.any():
        parity[v != 0] += 1
        v &= v - 1
    return parity


def mad_outliers(arr, cutoff=10, mad0="raise"):
    """Mask out mad outliers

    Mask out any values that are more than N times the median absolute
    devitation from the median.

    Although I (Gerrit Holl) came up with this myself, it's also
    documented at:

    http://eurekastatistics.com/using-the-median-absolute-deviation-to-find-outliers/

    except that I rolled by own approach for "what if mad==0".

    Note: If all values except one are constant, it is not possible to
    determine whether the remaining one is an outlier or “reasonably
    close” to the rest, without additional hints.  In this case, some
    outliers may go unnoticed.

    Arguments:

        arr (numpy.ndarray): n-D array with numeric dtype

        cutoff (int): Maximum tolerable normalised fractional distance

        mad0 (str): What to do if mad=0.  Can be 'raise', 'ignore', or
            'perc'.  In case of 'perc', will search for the lowest
            percentile at which the percentile absolute deviation is
            nonzero, increase the cutoff by the fractional approach toward
            percentile 100, and use that percentile instead.  So if the
            first non-zero is at percentile 75%, it will use the
            75th-percntile-absolute-deviation and increase the cutoff by
            a factor (100 - 50)/(100 - 75).

    Returns:

        ndarray with bool dtype, True for outliers
    """

    if arr.ptp() == 0:
        return np.zeros(shape=arr.shape, dtype="?")

    ad = abs(arr - np.ma.median(arr))
    mad = np.ma.median(ad)
    if mad == 0:
        if mad0 == "raise":
            raise ValueError("Cannot filter outliers, MAD=0")
        elif mad0 == "perc":
            # try other percentiles
            perc = np.r_[np.arange(50, 99, 1), np.linspace(99, 100, 100)]
            pad = scipy.stats.scoreatpercentile(ad, perc)
            if (pad == 0).all():  # all constant…?
                raise ValueError("These data are weird!")
            p_i = pad.nonzero()[0][0]
            cutoff *= (100 - 50) / (100 - perc[p_i])
            return (ad / pad[p_i]) > cutoff
    elif mad is np.ma.masked:
        # all are masked already…
        return np.ones(shape=ad.shape, dtype="?")
    else:
        return (ad / mad) > cutoff


def argclosest(array, value, retvalue=False):
    """Returns the index of the closest value in array.

    Parameters:
        array (ndarray): Input array.
        value (float): Value to compare to.
        retvalue (bool): If True, return the index and the closest value.

    Returns:
        int, float:
        Index of closest value, Closest value (if ``retvalue`` is True)

    """
    idx = np.abs(np.asarray(array) - value).argmin()

    return (idx, array[idx]) if retvalue else idx



"""Functions related to metrology

All functions in this module need sympy.
"""

# Any commits made to this module between 2015-05-01 and 2019-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import itertools
import warnings

def express_uncertainty(expr, aliases={}, on_failure="raise",
        collect_failures=None, return_sensitivities=False,
        return_components=False,
        correlated_terms=()):
    """For a sympy expression, calculate uncertainty.

    Uncertainty is given in the Guides to Uncertainties and Measurements
    (GUM), see http://www.bipm.org/en/publications/guides/gum.html
    equations 10 and 13.

    This takes all free symbols.  If you have IndexedBase symbols, you may
    want to pass them into aliases {free_symbol:
    indexed_base_symbol[index]} is this is not identified automatically.

    Arguments:
        
        expr (Expr): Expression for which to calculate uncertainty
        aliases (Mapping): Mapping of replacements to apply prior.
        on_failure (str): Signals what to do when some variables cannot be
            differentiated against.  This appears to be the case for
            indexed quantities (see
            https://github.com/sympy/sympy/issues/12191).  Default is
            'raise', but can set to 'warn' instead.
        collect_failures (set): Pass in a set object where failures will
            be collected, assuming on_failure is 'warn'.
        return_sensitivities (boolean): If True, return dictionary with
            sensitivity coefficients (as expressions).  Defaults to False.
        return_components (boolean): If True, return dictionary with
            component of evaluated uncertainty per argument/effect.
            Only considers uncorrelated part (which is the only thing
            implemented so far anyway).
        correlated_terms (Sequence[Symbol]): Sequence of symbols for which
            covariances will be included in the expression.  Passing a
            sequence with less than two elements means no covariances are
            considered.

    Returns:
        
        Expression indicating uncertainty
    """

    if collect_failures is None:
        collect_failures = set()
    import sympy
    u = sympy.Function("u")
    rv = sympy.sympify(0)
    sensitivities = {}
    components = {}
    recargs = recursive_args(expr,
            stop_at=(sympy.Symbol, sympy.Indexed,
                     sympy.concrete.expr_with_limits.ExprWithLimits))
    if recargs == set() and expr.args != (): # does not mean uncertainty is zero...!
        recargs.add(expr) # probably a loose Symbol or Indexed
    for sym in recargs:
        if isinstance(sym, sympy.concrete.expr_with_limits.ExprWithLimits):
            raise ValueError(f"Failed to express uncertainty in {expr!s}. "
                f"Uncertainty in {sym!s} not supported.  You must "
                "substitute limits and expand summation fully.")
        sym = aliases.get(sym, sym)
        try:
            sigma = sympy.diff(expr, sym)
            comp = sigma**2 * u(sym)**2
            rv += comp 
            sensitivities[sym] = sigma
            components[sym] = sympy.sqrt(comp)
        except ValueError as v:
            if on_failure == "raise" or not "derivative" in v.args[0]:
                raise
            else:
                warnings.warn("Unable to complete derivative on {!s}. "
                    "IGNORING uncertainty on {!s}!  Derivative failed with: "
                    "{:s}".format(expr, sym, v.args[0]))
                collect_failures.add(sym)
    # add cross-terms following Eq. 13 in the GUM
    for (sym_i, sym_j) in {x
            for x in itertools.product(correlated_terms, correlated_terms)
            if str(x[0])<str(x[-1])}:
        rv += 2 * sensitivities[sym_i] * sensitivities[sym_j] * u(sym_i, sym_j)
    unc = sympy.sqrt(rv)
    if return_sensitivities and return_components:
        return (unc, sensitivities, components)
    elif return_sensitivities:
        return (unc, sensitivities)
    elif return_components:
        return (unc, components)
    else:
        return unc

def recursive_args(expr, stop_at=None, partial_at=None):
    """Get arguments for `expr`, stopping at certain types

    Get all arguments in expression down to the levels in `expr`.  When
    `expr` is only `{sympy.Symbol}` this is identical to
    `expr.free_symbols`, but in some cases we want to retain IndexedBase,
    for example, when evaluating uncertainies.

    This is mainly a helper for express_uncertainty where we don't want to
    descend beyond Indexed quantities
    """

    import sympy
    import sympy.concrete.expr_with_limits
    if stop_at is None:
        stop_at = (sympy.Symbol, sympy.Indexed)
    if partial_at is None:
        partial_at = sympy.concrete.expr_with_limits.ExprWithLimits
    if isinstance(expr, stop_at):
        # to return {expr} here leads to infinite recursion!
        return set()
    if isinstance(expr, partial_at):
        return recursive_args(expr.args[0])
    args = set()
    for arg in expr.args:
        if isinstance(arg, stop_at):
            args.add(arg)
        elif isinstance(arg, partial_at):
            # in the case arg.args[0] is a stop_at, recursive_args will
            # return empty; but then we still need to add the expression a
            # level higher up.  For example, support we have
            # arg=Sum(T_PRT[n], (n, 0, N)), then arg.args[0]=T_PRT[n],
            # recursive_args(arg.args[0]) = set() (if sympy.Indexed is in
            # stop_at, such as by default), and neither gets added.
            # We should also make sure that any variable that is summed
            # over (stored in arg.args[1][0]) is excluded in any case.
            args.update((
                {arg.args[0]} if isinstance(arg.args[0], stop_at) else set()) |
                (recursive_args(arg.args[0], stop_at=stop_at)
                 - {arg.args[1][0]}))
        else:
            args.update(recursive_args(arg, stop_at=stop_at))
    return args



"""Functions related to colors."""
from .common import *

import matplotlib.colors as mcolors


TYPHON_COLORS = (
    ('jetblack', '#343434'),
    ('chalmers-blue', '#024a86'),
    ('chalmers-grey', '#b9cbcf'),
    ('darkgrey', '#767676'),
    ('eigengrau', '#16161d'),
    ('grey', '#b0b0b0'),
    ('max-planck', '#006c66'),
    ('uhh-red', '#ee1d23'),
    ('uhh-blue', '#00a1d7'),
    ('uhh-grey', '#314e58'),
)

# Add "ty:" prefix to prevent name collision with original matplotlib colors
TYPHON_COLORS = {'ty:' + name: value for name, value in TYPHON_COLORS}
mcolors.get_named_colors_mapping().update(TYPHON_COLORS)



# -*- coding: utf-8 -*-
"""Testing high-level functionality in typhon.arts.xml.
"""
from os.path import (dirname, join)

import numpy as np
import pytest

from typhon.arts import xml


class TestXML:
    """Testing high-level functionality in typhon.arts.xml."""
    ref_dir = join(dirname(__file__), "reference")

    def test_load_directory(self):
        """Test loading all XML files in a directory."""
        t = xml.load_directory(self.ref_dir)
        ref = xml.load(join(self.ref_dir, 'vector.xml'))

        assert np.allclose(t['vector'], ref)

    def test_load_directory_exclude(self):
        """Test excluding files when loading directory content."""
        t = xml.load_directory(self.ref_dir, exclude=['vector.xml'])

        with pytest.raises(KeyError):
            t['vector']



from os.path import dirname, join
import sys
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from typhon.collocations import collapse, Collocator, Collocations, expand
from typhon.files import FileSet, MHS_HDF
from typhon.files.utils import get_testfiles_directory
import xarray as xr


class TestCollocations:
    """Testing the collocation functions."""

    refdir = get_testfiles_directory("collocations")

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_search(self):
        """Collocate fake MHS filesets"""
        fake_mhs1 = FileSet(
            path=join(self.refdir,
                "{satname}_mhs_{year}", "{month}", "{day}", 
                "*NSS.MHSX.*.S{hour}{minute}.E{end_hour}{end_minute}.*.h5"),
            handler=MHS_HDF(),
        )
        fake_mhs2 = fake_mhs1.copy()

        with TemporaryDirectory() as outdir:
            collocations = Collocations(
                path=join(outdir, "{year}-{month}-{day}", 
                                  "{hour}{minute}{second}-{end_hour}{end_minute}{end_second}"),
            )
            collocations.search(
                [fake_mhs1, fake_mhs2],
                start="2007",
                end="2008", 
                max_interval="1h",
                max_distance="10km"
            )

    def test_flat_to_main_coord(self):
        """Tests Collocator._flat_to_main_coord

        This method is crucial since it stacks the whole input datasets for the
        collocating routine and makes them collocateable.
        """
        collocator = Collocator()

        test = xr.Dataset({
            "time": ("time", np.arange(10)),
            "lat": ("time", np.arange(10)),
            "lon": ("time", np.arange(10)),
        })
        check = xr.Dataset({
            "time": ("collocation", np.arange(10)),
            "lat": ("collocation", np.arange(10)),
            "lon": ("collocation", np.arange(10)),
        })
        results = collocator._flat_to_main_coord(test)
        assert check.equals(results)

        test = xr.Dataset({
            "time": ("main", np.arange(10)),
            "lat": ("main", np.arange(10)),
            "lon": ("main", np.arange(10)),
        })
        check = xr.Dataset({
            "time": ("collocation", np.arange(10)),
            "lat": ("collocation", np.arange(10)),
            "lon": ("collocation", np.arange(10)),
        })
        results = collocator._flat_to_main_coord(test)
        assert check.equals(results)

        test = xr.Dataset({
            "time": ("scnline", np.arange(5)),
            "lat": (("scnline", "scnpos"), np.arange(10).reshape(5, 2)),
            "lon": (("scnline", "scnpos"), np.arange(10).reshape(5, 2)),
        })
        check = test.stack(collocation=("scnline", "scnpos"))
        results = collocator._flat_to_main_coord(test)
        assert check.equals(results)

    def test_collocate_collapse_expand(self):
        """Test whether collocating, collapsing and expanding work"""
        collocator = Collocator()

        test = xr.Dataset({
            "time": ("time", np.arange("2000", "2010", dtype="M8[Y]")),
            "lat": ("time", np.arange(10)),
            "lon": ("time", np.arange(10)),
        })

        collocations = collocator.collocate(
            test, test, max_interval="30 days",
            max_distance="150 miles"
        )

        collapsed = collapse(collocations)
        expanded = expand(collocations)





# -*- coding: utf-8 -*-

"""Collection of utility functions."""

import numpy as np

from .xml.names import basic_types, tensor_names, complex_tensor_names


def get_arts_typename(var):
    """Returns the ARTS type name for this variable.

    Args:
        var: Variable to get the ARTS type name for.

    Returns:
        str: ARTS type name.

    """
    if type(var).__name__ in basic_types:
        ret = basic_types[type(var).__name__]
        if ret == 'Array':
            if len(var) == 0:
                return None
            else:
                element_type = get_arts_typename(var[0])
                for element in var[1:]:
                    if element_type != get_arts_typename(element):
                        return None
                ret = 'ArrayOf' + element_type
    elif isinstance(var, np.ndarray):
        if np.issubdtype(var.dtype, np.complex128):
            ret = complex_tensor_names[var.ndim - 1]
        else:
            ret = tensor_names[var.ndim - 1]
    else:
        ret = type(var).__name__

    return ret


def return_if_arts_type(var, artstype):
    """If variable is of specified ARTS type, return its value.

    Parameters:
        var: variable to check
        artstype: wanted ARTS type

    Returns: value of var, if var is specified type. None for NoneType
    """
    arts_name = get_arts_typename(var)
    if arts_name is None:
        return None
    elif arts_name == artstype:
        return var
    else:
        raise TypeError('Expected {} but got {}.' .format(artstype, arts_name))


def as_quantumnumbers(var):
    """Takes a quantum number prospect and turns it into a quantum number type
    if possible

    Parameters:
        var (dict, QuantumNumberRecord, QuantumNumbers, None, str):
            Quantum numbers

    Returns:
        QN (QuantumNumberRecord, QuantumNumbers): Returned quantumn numbers.
            No change if already quantum numbers type
    """

    if type(var) in [QuantumNumberRecord,
                     QuantumNumbers]:
        return var
    elif var is None:
        return QuantumNumbers()

    assert type(var) in [dict, str], "Cannot recognize as quantum number"

    if 'UP' in var and 'LO' in var:
        if type(var) is dict:
            return QuantumNumberRecord.from_dict(var)
        else:
            return QuantumNumberRecord.from_str(var)
    elif 'UP' in var:
        if type(var) is dict:
            var['LO'] = {}
            return QuantumNumberRecord.from_dict(var)
        else:
            return QuantumNumberRecord.from_str(var + ' LO')
    elif 'LO' in var:
        if type(var) is dict:
            var['UP'] = {}
            return QuantumNumberRecord.from_dict(var)
        else:
            return QuantumNumberRecord.from_str(var + ' UP')
    else:
        return QuantumNumbers(var)

from .catalogues import QuantumNumberRecord
from .catalogues import QuantumNumbers



"""Functions to estimate the different sources of retrieval error. """

from typhon.retrieval.oem import common


__all__ = [
    'smoothing_error',
    'retrieval_noise',
]


def smoothing_error(x, x_a, A):
    """Return the smoothing error through the averaging kernel.

    Parameters:
        x (ndarray): Atmospherice profile.
        x_a (ndarray): A priori profile.
        A (ndarray): Averaging kernel matrix.

    Returns:
        ndarray: Smoothing error due to correlation between layers.
    """
    return A @ (x - x_a)


def retrieval_noise(K, S_a, S_y, e_y):
    """Return the retrieval noise.

    Parameters:
        K (np.array): Simulated Jacobians.
        S_a (np.array): A priori error covariance matrix.
        S_y (np.array): Measurement covariance matrix.
        e_y (ndarray): Total measurement error.

    Returns:
        ndarray: Retrieval noise.
    """
    return common.retrieval_gain_matrix(K, S_a, S_y) @ e_y



"""
Utility functions.

Contains miscellaneous functions that do complicated things but
are note really relevant to other code.
"""

def unindent(source):
    """Unindent source code.

    Determines indent level of the first line and unindents
    all lines by the found indent level.

    Args:
        source: The source code to unindent as a string as
            obtained for example from inspect.getsource.

    Raises:
        Exception: If the non-whitespace characters are detected
            in the characters that are stripped off the code lines.

    Returns:
        new_source: The unindented source code.
    """

    if not type(source) == str:
        raise Exception("Argument must be a string.")

    lines = source.splitlines()

    if len(lines) < 1:
        return ""

    n_indent = len(lines[0]) - len(lines[0].lstrip())

    lines_new = []
    for i, l in enumerate(lines):
        stripped = l[:n_indent]
        if len(stripped.lstrip()) > 0:
            err = "Error when unindenting source code. Stripped characters" \
                  + stripped + " in line " + str(i) + " are non-whitespace " \
                  + " characters."
            raise Exception(err)
        lines_new += [l[n_indent:]]
    return "\n".join(lines_new)



# -*- coding: utf-8 -*-
"""Testing the covariance matrix class.
"""
import shutil

import pytest
import numpy as np
import scipy as sp
import os
from tempfile import mkstemp
from typhon.arts.covariancematrix import Block, CovarianceMatrix
from typhon.arts.xml import load, save

class TestCovarianceMatrix:

    def setup_method(self):
        # Temporary file
        fd, self.f = mkstemp()
        os.close(fd)

        # Simple covariance matrix for testing
        b1 = Block(0, 0, 0, 0, False, np.random.normal(size = (10, 10)))
        b2 = Block(1, 1, 10, 10, False, sp.sparse.identity(10))
        self.covmat = CovarianceMatrix([b1, b2])

    def test_xml_io(self):
        save(self.covmat, self.f)
        covmat2 = load(self.f)

        def compare_matrices(args):
            b1, b2 = args
            m1 = b1.matrix
            m2 = b2.matrix
            if isinstance(m1, sp.sparse.spmatrix):
                m1 = m1.todense()
                m2 = m2.todense()
                print(m1)
            return np.allclose(m1, m2)

        assert(all(map(compare_matrices, zip(self.covmat.blocks, covmat2.blocks))))

    def test_to_dense(self):
        m = self.covmat.to_dense()
        assert(np.allclose(m[:10, :10], self.covmat.blocks[0].matrix))
        assert(np.allclose(m[10:, 10:], self.covmat.blocks[1].matrix.toarray()))

    def teardown_method(self):
        # Remove temp file
        os.remove(self.f)



import os
from tempfile import TemporaryDirectory

from typhon.files import compress, decompress


class TestCompression:
    data = "ABCDEFGHIJKLMNOPQRSTUVWXYZ12345678910"

    def create_file(self, filename):
        with open(filename, "w") as file:
            file.write(self.data)

    def check_file(self, filename):
        with open(filename) as file:
            return self.data == file.readline()

    def test_compress_decompress_zip(self):
        with TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, 'testfile')
            with compress(tfile + ".zip") as compressed_file:
                self.create_file(compressed_file)

            with decompress(tfile + ".zip") as uncompressed_file:
                assert self.check_file(uncompressed_file)

    def test_compress_decompress_gzip(self):
        with TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, 'testfile')
            with compress(tfile + ".gz") as compressed_file:
                self.create_file(compressed_file)

            with decompress(tfile + ".gz") as uncompressed_file:
                assert self.check_file(uncompressed_file)

    def test_compress_decompress_bz2(self):
        with TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, 'testfile')
            with compress(tfile + ".bz2") as compressed_file:
                self.create_file(compressed_file)

            with decompress(tfile + ".bz2") as uncompressed_file:
                assert self.check_file(uncompressed_file)

    def test_compress_decompress_lzma(self):
        with TemporaryDirectory() as tdir:
            tfile = os.path.join(tdir, 'testfile')
            with compress(tfile + ".xz") as compressed_file:
                self.create_file(compressed_file)

            with decompress(tfile + ".xz") as uncompressed_file:
                assert self.check_file(uncompressed_file)



r"""
This module contains functions for the computations of scores for assessment
of retrieval methods.
"""
import numpy as np


def mape(y_pred, y_test):
    r"""
    The Mean Absolute Percentage Error (MAPE)

    The MAPE is computed as the mean of the absolute value of the relative
    error in percent, i.e.:

    .. math::

        \text{MAPE}(\mathbf{y}, \mathbf{y}_{true}) =
        \frac{100\%}{n}\sum_{i = 0}^n \frac{|y_{\text{pred},i} - y_{\text{true},i}|}
             {|y_{\text{true},i}|}

    Arguments:

        y_pred(numpy.array): The predicted scalar values.

        y_test(numpy.array): The true values.

    Returns:

        The MAPE for the given predictions.

    """
    return np.nanmean(100.0 * np.abs(y_test - y_pred.ravel()) / np.abs(y_test).ravel())


def bias(y_pred, y_test):
    r"""
    The mean bias in percent.

    .. math::

        \text{BIAS}(\mathbf{y}, \mathbf{y}_{true}) =
        \frac{100\%}{n}\sum_{i = 0}^n \frac{y_{\text{pred},i} - y_{\text{true},i}}
             {y_{\text{true},i}}

    Arguments:

        y_pred(numpy.array): The predicted scalar values.

        y_test(numpy.array): The true values.

    Returns:

        The mean bias in percent.

    """
    return np.mean(100.0 * y_test - y_pred / y_test)


def quantile_score(y_tau, y_test, taus):
    r"""
    The quantile loss function.

    Let $y_\tau$ be an estimate of the $\tau\text{th}$ quantile of a
    random distribution $Y$. Then the quantile loss is a proper scoring
    rule that measures the quality of the estimate $y_\tau$ given a
    sample $y$ from $Y$:

    .. math::
            \mathcal{L}_\tau(y_\tau, y_{true}) =
            \begin{cases} (1 - \tau)|y_\tau - y_{true}| & \text{ if } y_\tau < y_\text{true} \\
            \tau |y_\tau - y_\text{true}| & \text{ otherwise. }\end{cases}

    Arguments:

        y_tau(numpy.array): Numpy array with shape (n, k) containing one row of
                            k estimated quantiles for each of the n test cases.

        y_test(numpy.array): Numpy array with the n observed test values
                             of the conditional distributions whose quantiles are
                             estimated by the elements in `y_tau`

        taus(numpy.array): Numpy array containing the k quantile fractions
                           :math:`\tau` that are estimated by the columns in
                           `y_tau`.

    Returns:

        Array of shape (n, k) containing the quantile scores for each quantile
        estimate.

    Raises:

        ValueError
            If the shapes of `y_tau`, `y_test` and `taus` are inconsistent.
    """
    taus = np.asarray(taus)
    m = taus.size

    y_tau = y_tau.reshape(-1, m)
    n = y_tau.shape[0]

    try:
        y_test = y_test.reshape(n, 1)
    except:
        raise ValueError(
            "Shape of y_test is incompatible with y_tau and taus.")

    abs_1 = taus * np.abs(y_tau - y_test)
    abs_2 = (1.0 - taus) * np.abs(y_tau - y_test)

    return np.where(y_tau < y_test, abs_1, abs_2)


def mean_quantile_score(y_tau, y_test, taus):
    r"""
    Wrapper around the `quantile_score` function, which computes the mean
    along the first dimension.
    """
    return np.nanmean(quantile_score(y_tau, y_test, taus), axis=0)



# -*- coding: utf-8 -*-

"""Module for anything related to the electromagnetic spectrum.
"""
import numpy as np

from typhon import constants


__all__ = [
    'planck',
    'planck_wavelength',
    'planck_wavenumber',
    'rayleighjeans',
    'rayleighjeans_wavelength',
    'radiance2planckTb',
    'radiance2rayleighjeansTb',
    'snell',
    'fresnel',
    'frequency2wavelength',
    'frequency2wavenumber',
    'wavelength2frequency',
    'wavelength2wavenumber',
    'wavenumber2frequency',
    'wavenumber2wavelength',
    'stefan_boltzmann_law',
    'zeeman_splitting',
    'zeeman_strength',
    'zeeman_transitions',
    ]


def planck(f, T):
    """Calculate black body radiation for given frequency and temperature.

    Parameters:
        f (float or ndarray): Frquencies [Hz].
        T (float or ndarray): Temperature [K].

    Returns:
        float or ndarray: Radiances.

    """
    c = constants.speed_of_light
    h = constants.planck
    k = constants.boltzmann

    return 2 * h * f**3 / (c**2 * (np.exp(np.divide(h * f, (k * T))) - 1))


def planck_wavelength(l, T):
    """Calculate black body radiation for given wavelength and temperature.

    Parameters:
        l (float or ndarray): Wavelength [m].
        T (float or ndarray): Temperature [K].

    Returns:
        float or ndarray: Radiances.

    """
    c = constants.speed_of_light
    h = constants.planck
    k = constants.boltzmann

    return 2 * h * c**2 / (l**5 * (np.exp(np.divide(h * c, (l * k * T))) - 1))


def planck_wavenumber(n, T):
    """Calculate black body radiation for given wavenumber and temperature.

    Parameters:
        n (float or ndarray): Wavenumber.
        T (float or ndarray): Temperature [K].

    Returns:
        float or ndarray: Radiances.

    """
    c = constants.speed_of_light
    h = constants.planck
    k = constants.boltzmann

    return 2 * h * c**2 * n**3 / (np.exp(np.divide(h * c * n, (k * T))) - 1)


def rayleighjeans(f, T):
    """Calculates the Rayleigh-Jeans approximation of the Planck function.

     Calculates the approximation of the Planck function for given
     frequency and temperature.

     Parameters:
        f (float or ndarray): Frequency [Hz].
        T (float or ndarray): Temperature [K].

     Returns:
        float or ndarray: Radiance [W/(m2*Hz*sr)].

    """
    c = constants.speed_of_light
    k = constants.boltzmann

    return 2 * f**2 * k * T / c**2


def rayleighjeans_wavelength(l, T):
    """Calculates the Rayleigh-Jeans approximation of the Planck function.

     Calculates the approximation of the Planck function for given
     wavelength and temperature.

     Parameters:
        l (float or ndarray): Wavelength [m].
        T (float or ndarray): Temperature [K].

     Returns:
        float or ndarray: Radiance [W/(m2*Hz*sr)].

    """
    c = constants.speed_of_light
    k = constants.boltzmann

    return np.divide(2 * c * k * T, l**4)


def radiance2planckTb(f, r):
    """Convert spectral radiance to Planck brightness temperture.

    Parameters:
        f (float or ndarray): Frequency [Hz].
        r (float or ndarray): Spectral radiance [W/m**-2/sr].

    Returns:
        float or ndarray: Planck brightness temperature [K].
    """
    c = constants.speed_of_light
    k = constants.boltzmann
    h = constants.planck

    return h / k * f / np.log(np.divide((2 * h / c**2) * f**3, r) + 1)


def radiance2rayleighjeansTb(f, r):
    """Convert spectral radiance to Rayleight-Jeans brightness temperture.

    Parameters:
        f (float or ndarray): Frequency [Hz].
        r (float or ndarray): Spectral radiance [W/m**-2/sr].

    Returns:
        float or ndarray: Rayleigh-Jeans brightness temperature [K].
    """
    c = constants.speed_of_light
    k = constants.boltzmann

    return np.divide(c**2, (2 * f**2 * k)) * r


def snell(n1, n2, theta1):
    """Calculates the angle of the transmitted wave, according to Snell's law.

    Snell's law for the case when both *n1* and *n2* have no imaginary part
    is found in all physics handbooks.

    The expression for complex *n2* is taken from "An introduction to
    atmospheric radiation" by K.N. Liou (Sec. 5.4.1.3).

    No expression that allows *n1* to be complex has been found.

    If theta2 is found to be complex, it is returned as NaN. This can happen
    when n1 > n2, and corresponds to a total reflection and there is no
    transmitted part.

    The refractive index and the dielectric constant, epsilon, are releated
    as

    .. math::
        n = \sqrt{\epsilon}

    Parameters:
        n1 (float or ndarray): Refractive index for medium of incoming
            radiation.
        n2 (float or ndarray): Refractive index for reflecting medium.
        theta1 (float): Angle between surface normal
            and incoming radiation [degree].

    Returns:
        float or ndarray: Angle for transmitted part [degree].

    .. Ported from atmlab. Original author: Patrick Eriksson
    """

    if np.any(np.real(n1) <= 0) or np.any(np.real(n2) <= 0):
        raise Exception('The real part of *n1* and *n2* can not be <= 0.')

    if np.all(np.isreal(n1)) and np.all(np.isreal(n2)):
        theta2 = np.arcsin(n1 * np.sin(np.deg2rad(theta1)) / n2)

    elif np.all(np.isreal(n1)):
        mr2 = (np.real(n2) / n1)**2
        mi2 = (np.imag(n2) / n1)**2
        sin1 = np.sin(np.deg2rad(theta1))
        s2 = sin1 * sin1
        Nr = np.sqrt((mr2 - mi2 + s2
                      + np.sqrt((mr2 - mi2 - s2)**2 + 4 * mr2 * mi2)
                      ) / 2)
        theta2 = np.arcsin(sin1 / Nr)
    else:
        raise Exception('No expression implemented for imaginary *n1*.')

    if not np.all(np.isreal(theta2)):
        theta2 = np.nan

    return np.rad2deg(theta2)


def fresnel(n1, n2, theta1):
    r"""Fresnel formulas for surface reflection.

    The amplitude reflection coefficients for a flat surface can also be
    calculated (Rv and Rh). Note that these are the coefficients for the
    amplitude of the wave. The power reflection coefficients are
    obtained as

    .. math::
        r = \lvert R \rvert^2

    The expressions used are taken from Eq. 3.31 in "Physical principles of
    remote sensing", by W.G. Rees, with the simplification that that relative
    magnetic permeability is 1 for both involved media. The theta2 angle is
    taken from snell.m.

    The refractive index of medium 2  (n2) can be complex. The refractive
    index and the dielectric constant, epsilon, are releated as

    .. math::
        n = \sqrt{\epsilon}

    No expression for theta2 that allows *n1* to be complex has been found.

    If theta2 is found to be complex, it is returned as NaN. This can happen
    when n1 > n2, and corresponds to a total reflection and there is no
    transmitted part. Rv and Rh are here set to 1.

    Parameters:
        n1 (float or ndarray): Refractive index for medium of incoming
            radiation.
        n2 (float or ndarray): Refractive index for reflecting medium.
        theta1 (float or ndarray): Angle between surface normal
            and incoming radiation [degree].

    Returns:
        float or ndarray, float or ndarray:
            Reflection coefficient for vertical polarisation,
            reflection coefficient for horisontal polarisation.

    .. Ported from atmlab. Original author: Patrick Eriksson
    """
    if np.any(np.imag(n1) < 0) or np.any(np.imag(n2) < 0):
        raise Exception(
            'The imaginary part of *n1* and *n2* can not be negative.')

    theta2 = snell(n1, n2, theta1)

    costheta1 = np.cos(np.deg2rad(theta1))
    costheta2 = np.cos(np.deg2rad(theta2))

    Rv = (n2 * costheta1 - n1 * costheta2) / (n2 * costheta1 + n1 * costheta2)
    Rh = (n1 * costheta1 - n2 * costheta2) / (n1 * costheta1 + n2 * costheta2)

    return Rv, Rh


def frequency2wavelength(frequency):
    """Convert frequency to wavelength.

    Parameters:
        frequency (float or ndarray): Frequency [Hz].

    Returns:
        float or ndarray: Wavelength [m].

    """
    return np.divide(constants.speed_of_light, frequency)


def frequency2wavenumber(frequency):
    """Convert frequency to wavenumber.

    Parameters:
        frequency (float or ndarray): Frequency [Hz].

    Returns:
        float or ndarray: Wavenumber [m^-1].

    """
    return frequency / constants.speed_of_light


def wavelength2frequency(wavelength):
    """Convert wavelength to frequency.

    Parameters:
        wavelength (float or ndarray): Wavelength [m].

    Returns:
        float or ndarray: Frequency [Hz].

    """
    return np.divide(constants.speed_of_light, wavelength)


def wavelength2wavenumber(wavelength):
    """Convert wavelength to wavenumber.

    Parameters:
        wavelength (float or ndarray): Wavelength [m].

    Returns:
        float or ndarray: Wavenumber [m^-1].

    """
    return np.divide(1, wavelength)


def wavenumber2frequency(wavenumber):
    """Convert wavenumber to frequency.

    Parameters:
        wavenumber (float or ndarray): Wavenumber [m^-1].
    Returns:
        float or ndarray: Frequency [Hz].

    """
    return constants.speed_of_light * wavenumber


def wavenumber2wavelength(wavenumber):
    """Convert wavenumber to wavelength.

    Parameters:
        wavenumber (float or ndarray): Wavenumber [m^-1].

    Returns:
        float or ndarray: Wavelength [m].

    """
    return np.divide(1, wavenumber)


def stefan_boltzmann_law(T):
    """Compute Stefan Boltzmann law for given temperature

    .. math::
        j = \\sigma T^4

    Parameters:
        T (float or ndarray): Physical temperature  of object [K]

    Returns:
        float or ndarray: Energy per surface area [W m^-2]
    """
    return constants.stefan_boltzmann_constant * T**4


def hund_case_a_landau_g_factor(o, j, s, l, gs, gl):
    """ Hund case A Landau g-factor

    .. math::
        g =  \\frac{\\Omega}{J(J+1)} \\left( g_sS + g_l\\Lambda \\right)

    Parameters:
        o (float): Omega of level

        j (float): J of level

        l (float): Lambda of level

        s (float): Sigma of level

        gs (float): relativistic spin factor of molecule

        gl (float): relativistic orbital factor of molecule

    Returns:
        g (float): Landau g-factor for Hund case A
    """
    return gs * o * s / (j * (j + 1)) + gl * o * l / (j * (j + 1))


def hund_case_b_landau_g_factor(n, j, s, l, gs, gl):
    """ Hund case B Landau g-factor

    .. math::
        g = g_s \\frac{J(J+1) + S(S+1) - N(N+1)}{2J(J+1)} +
        g_l \\Lambda \\frac{J(J+1) - S(S+1) + N(N+1)}{2JN(J+1)(N+1)}

    Parameters:
        n (float): N of level

        j (float): J of level

        l (float): Lambda of level

        s (float): S of level

        gs (float): relativistic spin factor of molecule

        gl (float): relativistic orbital factor of molecule

    Returns:
        g (float): Landau g-factor for Hund case B
    """
    if n != 0:
        return gs * (j * (j+1) + s * (s+1) - n * (n+1)) / (2 * j * (j + 1)) + \
            l * gl * (j * (j+1) - s * (s+1) + n * (n+1)) / (2 * j * n * (j + 1) * (n + 1))
    else:
        return gs * (j * (j+1) + s * (s+1) - n * (n+1)) / (2 * j * (j + 1))


def landau_g_factor(x, j, s, l=None, gs=2, gl=1, case=None):
    """ The Landau g-factor

    Parameters:
        x (float): N or Omega of level

        j (float): J of level

        s (float): S of level

        l (float): Lambda of level

        gs (float): relativistic spin factor of molecule

        gl (float): relativistic orbital factor of molecule

        case (str): Case type, 'a' or 'b'

    Returns:
        g (float): Landau g-factor for the Hund case or None for unknown case
    """
    if j == 0:
        return 0
    elif l is None:
        l = 0

    assert type(case) == str, "Case must be str for J not 0"

    if case.lower() == 'a':
        return hund_case_a_landau_g_factor(x, j, s, l, gs, gl)
    elif case.lower() == 'b':
        return hund_case_b_landau_g_factor(x, j, s, l, gs, gl)
    else:
        raise RuntimeError("Unknown case-type: " + str(case))


def zeeman_splitting(gu, gl, mu, ml, H=1):
    """ Zeeman splitting

    .. math::
        \Delta f = \\frac{H\mu_b}{h}(g_um_u - g_lm_l),

    where :math:`\mu_b` is the Bohr magneton and :math:`h` is the Planck
    constant.

    Parameters:
        gu (scalar or ndarray): Upper g

        gl (scalar or ndarray): Lower g

        mu (scalar or ndarray): Upper projection of j

        ml (scalar or ndarray): Lower projection of j

        H (scalar or ndarray): Absolute strength of magnetic field in Teslas

    Returns:
        scalar or ndarray: Splitting in Hertz
    """
    h = constants.planck
    mu_B = constants.mu_B

    frac = mu_B / h
    return frac * (gu * mu - gl * ml) * H


def zeeman_strength(ju, jl, mu, ml):
    """ Zeeman line strength

    .. math:: \Delta S_{M_u,M_l} = C \\left(\\begin{array}{ccc} J_l & 1 & J_u
                       \\\\ M_l & M_u-M_l&-M_u \\end{array}\\right)^2,

    where C is either 3/2 or 3/4 depending on in mu-ml is 0 or not.  In case
    the intent is to return many split lines, the size of either (jl, ju) or
    (ml, mu) must be the same.  Regardless, these pairs have to be of the same
    type

    Parameters:
        ju (scalar or 1darray): Upper level J.  Must be same size as jl

        jl (scalar or 1darray): Lower level J.  Must be same size as ju

        mu (scalar or ndarray): Upper level M.  Must be same size as ml unless
        J is vector-type, then it is ignored

        ml (scalar or ndarray): Lower level M.  Must be same size as mu unless
        J is vector-type, then it is ignored

    Returns:
        scalar or ndarray: Relative line strength of component normalized to 1
        or array(array(S+, Pi, S-))
    """
    assert type(jl) == type(ju), "Must have same type: J"
    assert type(ml) == type(mu), "Must have same type: M"

    try:
        import sympy.physics.wigner as wig
    except ModuleNotFoundError:
        raise RuntimeError("Must have sympy installed to use")

    if np.isscalar(mu) and np.isscalar(ju):
        dm = mu - ml
        w = wig.wigner_3j(jl, 1, ju, ml, dm, -mu)
        w = float(w)
        if dm == 0:
            r = w**2 * 1.5
        else:
            r = w**2 * 0.75
    elif np.isscalar(ju):
        r = []
        for i in range(len(mu)):
            r.append(zeeman_strength(ju, jl, mu[i], ml[i]))
        r = np.array(r)
    else:
        r = []
        for i in range(len(ju)):
            JU = ju[i]
            JL = jl[i]
            if np.isscalar(JU):
                t = []
                MU, ML = zeeman_transitions(JU, JL, "S-")
                t.append(zeeman_strength(JU, JL, MU, ML))
                MU, ML = zeeman_transitions(JU, JL, "Pi")
                t.append(zeeman_strength(JU, JL, MU, ML))
                MU, ML = zeeman_transitions(JU, JL, "S+")
                t.append(zeeman_strength(JU, JL, MU, ML))
                r.append(t)
            else:
                r.append(zeeman_strength(JU, JL, mu, ml))
        r = np.array(r)
    return r


def zeeman_transitions(ju, jl, type):
    """ Find possible mu and ml for valid ju and jl for a given transistion
    polarization

    Parameters:
        ju (scalar):  Upper level J

        jl (scalar):  Lower level J

        type (string): "Pi", "S+", or "S-" for relevant polarization type

    Returns:
        tuple: MU, ML arrays for given Js and polarization type
    """
    assert np.isscalar(ju) and np.isscalar(jl), "non-scalar J non supported"
    assert type.lower() in ["pi", "s+", "s-"], "unknown transition type"
    assert ju - jl in [-1, 0, 1], "delta-J should belong to {-1, 0, 1}"
    assert ju > 0 and jl >= 0, "only for positive ju and non-negative for jl"

    if type.lower() == "pi":
        J = min(ju, jl)
        return np.arange(-J, J + 1), np.arange(-J, J + 1)
    elif type.lower() == "s+":
        if ju < jl:
            return np.arange(-ju, ju+1), np.arange(-ju+1, ju+2)
        elif ju == jl:
            return np.arange(-ju, ju), np.arange(-ju+1, ju+1)
        else:
            return np.arange(-ju, jl), np.arange(-ju+1, jl+1)
    elif type.lower() == "s-":
        if ju < jl:
            return np.arange(-ju, ju+1), np.arange(-jl, ju)
        elif ju == jl:
            return np.arange(-ju+1, ju+1), np.arange(-ju, ju)
        else:
            return np.arange(-ju+2, ju+1), np.arange(-ju+1, ju)


def zeeman_theta(u, v, w, z=0, a=0):
    """ Find Zeeman angle along the magnetic field
    """

    try:
        import sympy as sp
    except ModuleNotFoundError:
        raise RuntimeError("Must have sympy installed to use")

    U, V, W, Z, A = np.meshgrid(u, v, w, z, a, copy=False)
    N = len(U.flatten())

    if type(sp.symbols('u')) == type(u):
        sin = sp.sin
        cos = sp.cos
        acos = sp.acos
        sqrt = sp.sqrt
        d = np.empty((N), type(u))
    else:
        sin = np.sin
        cos = np.cos
        acos = np.arccos
        sqrt = np.sqrt
        d = np.empty((N), float)

    for i in range(N):
        H = np.array([U.flat[i], V.flat[i], W.flat[i]])
        L = np.array([sin(Z.flat[i])*cos(A.flat[i]),
                      sin(Z.flat[i])*sin(A.flat[i]), cos(Z.flat[i])])
        d[i] = acos(H.dot(L) / sqrt(H.dot(H)))

    if type(sp.symbols('u')) == type(u):
        return d[0]

    shape = []
    for input in [u, v, w, z, a]:
        if np.isscalar(input):
            continue
        shape.append(len(input))

    if shape:
        d = d.reshape(shape)
    else:
        d = d[0]

    return d


def zeeman_eta(u, v, w, z=0, a=0):

    """ Find Zeeman angle along the magnetic field
    """

    try:
        import sympy as sp
    except ModuleNotFoundError:
        raise RuntimeError("Must have sympy installed to use")

    U, V, W, Z, A = np.meshgrid(u, v, w, z, a, copy=False)
    N = len(U.flatten())

    if type(sp.symbols('u')) == type(u):
        sin = sp.sin
        cos = sp.cos
        atan2 = sp.atan2
        d = np.empty((N), dtype=type(u))
    else:
        sin = np.sin
        cos = np.cos
        atan2 = np.arctan2
        d = np.empty((N), dtype=float)

    for i in range(N):
        H = np.array([U.flat[i], V.flat[i], W.flat[i]])
        L = np.array([sin(Z.flat[i])*cos(A.flat[i]),
                      sin(Z.flat[i])*sin(A.flat[i]), cos(Z.flat[i])])
        R = np.array([-sin(A.flat[i]), cos(A.flat[i]), 0])
        p = H - L.dot(H) * H
        d[i] = atan2(np.cross(R, p).dot(L), p.dot(R))

    if type(sp.symbols('u')) == type(u):
        return d[0]

    shape = []
    for input in [u, v, w, z, a]:
        if np.isscalar(input):
            continue
        shape.append(len(input))

    if shape:
        d = d.reshape(shape)
    else:
        d = d[0]

    return d
